"""Base class for selection pipelines."""

from copy import copy
from copy import deepcopy
from typing import Any
from typing import List
from typing import Optional
from typing import Sequence
from typing import Tuple
from typing import Union

from pandas import Series

from lightautoml.validation.base import TrainValidIterator

from ...dataset.base import LAMLDataset
from ...ml_algo.base import MLAlgo
from ...ml_algo.tuning.base import DefaultTuner
from ...ml_algo.tuning.base import ParamsTuner
from ...ml_algo.utils import tune_and_fit_predict
from ..features.base import FeaturesPipeline
from ..utils import map_pipeline_names


class ImportanceEstimator:
    """Abstract class, that estimates feature importances."""

    def __init__(self):
        self.raw_importances = None

    # Change signature here to be compatible with MLAlgo
    def fit(self, *args: Any, **kwargs: Any):
        """Calculate feature importance."""
        raise NotImplementedError

    def get_features_score(self) -> Series:
        """Get raw features importances.

        Returns:
            Pandas Series object with index - str features names and values - array of importances.

        """
        return self.raw_importances


class SelectionPipeline:
    """Abstract class, performing feature selection.

    Instance should accept train/valid datasets and select features.

    Args:
        features_pipeline: Composition of feature transforms.
        ml_algo: Tuple (MlAlgo, ParamsTuner).
        imp_estimator: Feature importance estimator.
        fit_on_holdout: If use the holdout iterator.
        **kwargs: Not used.

    """

    @property
    def is_fitted(self) -> bool:
        """Check if selection pipeline is already fitted.

        Returns:
            ``True`` for fitted pipeline and False for not fitted.

        """
        return self._selected_features is not None

    @property
    def selected_features(self) -> List[str]:
        """Get selected features.

        Returns:
            List of selected feature names.

        """
        assert self._selected_features is not None, "Should be fitted first"
        forced_features = self.forced_features if self.forced_features is not None else []
        return list(set(self._selected_features + forced_features))

    @selected_features.setter
    def selected_features(self, val: List[str]):
        """Setter of selected features.

        Args:
            val: List of selected feature names.

        """
        self._selected_features = deepcopy(val)

    @property
    def in_features(self) -> List[str]:
        """Input features to the selector.

        Raises exception if not fitted beforehand.

        Returns:
            List of input features.

        """
        assert self._in_features is not None, "Should be fitted first"
        return self._in_features

    @property
    def dropped_features(self) -> List[str]:
        """Features that were dropped.

        Returns:
            list of dropped features.

        """
        included = set(self._selected_features)
        return [x for x in self._in_features if x not in included]

    def __init__(
        self,
        features_pipeline: Optional[FeaturesPipeline] = None,
        ml_algo: Optional[Union[MLAlgo, Tuple[MLAlgo, ParamsTuner]]] = None,
        imp_estimator: Optional[ImportanceEstimator] = None,
        fit_on_holdout: bool = False,
        **kwargs: Any
    ):
        self.features_pipeline = features_pipeline
        self._fit_on_holdout = fit_on_holdout

        self.ml_algo = None
        self._empty_algo = None
        if ml_algo is not None:
            try:
                self.ml_algo, self.tuner = ml_algo
            except (TypeError, ValueError):
                self.ml_algo, self.tuner = ml_algo, DefaultTuner()

            if not self.ml_algo.is_fitted:
                self._empty_algo = deepcopy(self.ml_algo)

        self.imp_estimator = imp_estimator
        self._selected_features = None
        self._in_features = None
        self.mapped_importances = None
        self.forced_features = None

    def perform_selection(self, train_valid: Optional[TrainValidIterator]):
        """Select features from train-valid iterator.

        Method is used to perform selection based
        on features pipeline and ml model.
        Should save ``_selected_features`` attribute in the end of working.

        Args:
            train_valid: Classical cv-iterator.

        Raises:
            NotImplementedError.

        """
        raise NotImplementedError

    def fit(self, train_valid: TrainValidIterator):
        """Selection pipeline fit.

        Find features selection for given dataset based
        on features pipeline and ml model.

        Args:
            train_valid: Dataset iterator.

        """
        if not self.is_fitted:

            if self._fit_on_holdout:
                train_valid = train_valid.convert_to_holdout_iterator()

            self._in_features = train_valid.features
            if self.features_pipeline is not None:
                train_valid = train_valid.apply_feature_pipeline(self.features_pipeline)

            preds = None
            if self.ml_algo is not None:
                if self.ml_algo.is_fitted:
                    assert list(self.ml_algo.features) == list(
                        train_valid.features
                    ), "Features in feated MLAlgo should match exactly"
                else:
                    self.ml_algo, preds = tune_and_fit_predict(self.ml_algo, self.tuner, train_valid)

            if self.imp_estimator is not None:
                self.imp_estimator.fit(train_valid, self.ml_algo, preds)

            self.perform_selection(train_valid)

    def select(self, dataset: LAMLDataset) -> LAMLDataset:
        """Takes only selected features from giving dataset and creates new dataset.

        Args:
            dataset: Dataset for feature selection.

        Returns:
            New dataset with selected features only.

        """
        selected_features = copy(self.selected_features)

        # Add forced features
        if self.forced_features is None:
            self.forced_features = [feature for feature in dataset.features if dataset.roles[feature].force_input]

        return dataset[:, set(selected_features + self.forced_features)]

    def map_raw_feature_importances(self, raw_importances: Series):
        """Calculate input feature importances.

        Calculated as sum of importances on different levels of pipeline.

        Args:
            raw_importances: Importances of output features.

        # noqa: DAR201

        """
        if self.features_pipeline is None:
            return raw_importances.copy()
        mapped = map_pipeline_names(self.in_features, raw_importances.index)
        mapped_importance = Series(raw_importances.values, index=mapped)

        self.mapped_importances = mapped_importance.groupby(level=0).sum().sort_values(ascending=False)

    def get_features_score(self):
        """Get input feature importances.

        Returns:
            Series with importances in not ascending order.

        """
        return self.mapped_importances


class EmptySelector(SelectionPipeline):
    """Empty selector - perform no selection, just save input features names."""

    def __init__(self):
        super().__init__()

    def perform_selection(self, train_valid: Optional[TrainValidIterator]):
        """Just save input features names.

        Args:
            train_valid: Used for getting features names.

        """
        self._selected_features = train_valid.features


class PredefinedSelector(SelectionPipeline):
    """Predefined selector - selects columns specified by user.

    Args:
        columns_to_select: Columns will be selected.

    """

    def __init__(self, columns_to_select: Sequence[str]):
        super().__init__()
        self.columns_to_select = set(columns_to_select)

    def perform_selection(self, train_valid: Optional[TrainValidIterator]):
        """Select only specified columns.

        Args:
            train_valid: Used for validation of features presence.

        """
        assert len(self.columns_to_select) == len(
            self.columns_to_select.intersection(set(train_valid.features))
        ), "Columns to select not match with dataset features"
        self._selected_features = sorted(list(self.columns_to_select))


class ComposedSelector(SelectionPipeline):
    """Composed selector - perform composition of selections.

    Args:
        selectors: Sequence of selectors.

    """

    def __init__(self, selectors: Sequence[SelectionPipeline]):
        super().__init__()
        self.selectors = selectors

    def fit(self, train_valid: Optional[TrainValidIterator] = None):
        """Fit all selectors in composition.

        Args:
            train_valid: Dataset iterator.

        """
        for selector in self.selectors:
            train_valid = train_valid.apply_selector(selector)

        self._in_features = self.selectors[0].in_features
        self.perform_selection(train_valid)

    def perform_selection(self, train_valid: Optional[TrainValidIterator]):
        """Defines selected features.

        Args:
            train_valid: Not used.

        """
        self._selected_features = self.selectors[-1].selected_features

    def get_features_score(self):
        """Get mapped input features importances."""
        return self.selectors[-1].mapped_importances



"""Utils for pytorch-tabnet model."""
import torch
import numpy as np
import torch.nn as nn
from ..torch_based.node_nn_model import Entmax15, Sparsemax
from ..torch_based.autoint.ghost_norm import GhostBatchNorm


def _initialize_non_glu(module, input_dim, output_dim):
    gain_value = np.sqrt((input_dim + output_dim) / np.sqrt(4 * input_dim))
    torch.nn.init.xavier_normal_(module.weight, gain=gain_value)
    # torch.nn.init.zeros_(module.bias)
    return


def _initialize_glu(module, input_dim, output_dim):
    gain_value = np.sqrt((input_dim + output_dim) / np.sqrt(input_dim))
    torch.nn.init.xavier_normal_(module.weight, gain=gain_value)
    # torch.nn.init.zeros_(module.bias)
    return


class TabNetEncoder(torch.nn.Module):
    """Defines main part of the TabNet network without the embedding layers.

    Code from https://github.com/dreamquark-ai/tabnet

    Args:
        input_dim : int
            Number of features
        output_dim : int or list of int for multi task classification
            Dimension of network output
            examples : one for regression, 2 for binary classification etc...
        n_d : int
            Dimension of the prediction  layer (usually between 4 and 64)
        n_a : int
            Dimension of the attention  layer (usually between 4 and 64)
        n_steps : int
            Number of successive steps in the network (usually between 3 and 10)
        gamma : float
            Float above 1, scaling factor for attention updates (usually between 1.0 to 2.0)
        n_independent : int
            Number of independent GLU layer in each GLU block (default 2)
        n_shared : int
            Number of independent GLU layer in each GLU block (default 2)
        epsilon : float
            Avoid log(0), this should be kept very low
        virtual_batch_size : int
            Batch size for Ghost Batch Normalization
        momentum : float
            Float value between 0 and 1 which will be used for momentum in all batch norm
        mask_type : str
            Either "sparsemax" or "entmax" : this is the masking function to use
        group_attention_matrix : torch matrix
            Matrix of size (n_groups, input_dim), m_ij = importance within group i of feature j
    """

    def __init__(
        self,
        input_dim,
        output_dim,
        n_d=8,
        n_a=8,
        n_steps=3,
        gamma=1.3,
        n_independent=2,
        n_shared=2,
        epsilon=1e-15,
        virtual_batch_size=128,
        momentum=0.02,
        mask_type="sparsemax",
        group_attention_matrix=None,
    ):
        super(TabNetEncoder, self).__init__()
        self.input_dim = input_dim
        self.output_dim = output_dim
        self.is_multi_task = isinstance(output_dim, list)
        self.n_d = n_d
        self.n_a = n_a
        self.n_steps = n_steps
        self.gamma = gamma
        self.epsilon = epsilon
        self.n_independent = n_independent
        self.n_shared = n_shared
        self.virtual_batch_size = virtual_batch_size
        self.mask_type = mask_type
        self.initial_bn = nn.BatchNorm1d(self.input_dim, momentum=0.01)
        self.group_attention_matrix = group_attention_matrix

        if self.group_attention_matrix is None:
            # no groups
            self.group_attention_matrix = torch.eye(self.input_dim)
            self.attention_dim = self.input_dim
        else:
            self.attention_dim = self.group_attention_matrix.shape[0]

        if self.n_shared > 0:
            shared_feat_transform = torch.nn.ModuleList()
            for i in range(self.n_shared):
                if i == 0:
                    shared_feat_transform.append(nn.Linear(self.input_dim, 2 * (n_d + n_a), bias=False))
                else:
                    shared_feat_transform.append(nn.Linear(n_d + n_a, 2 * (n_d + n_a), bias=False))

        else:
            shared_feat_transform = None

        self.initial_splitter = FeatTransformer(
            self.input_dim,
            n_d + n_a,
            shared_feat_transform,
            n_glu_independent=self.n_independent,
            virtual_batch_size=self.virtual_batch_size,
            momentum=momentum,
        )

        self.feat_transformers = torch.nn.ModuleList()
        self.att_transformers = torch.nn.ModuleList()

        for step in range(n_steps):
            transformer = FeatTransformer(
                self.input_dim,
                n_d + n_a,
                shared_feat_transform,
                n_glu_independent=self.n_independent,
                virtual_batch_size=self.virtual_batch_size,
                momentum=momentum,
            )
            attention = AttentiveTransformer(
                n_a,
                self.attention_dim,
                group_matrix=group_attention_matrix,
                virtual_batch_size=self.virtual_batch_size,
                momentum=momentum,
                mask_type=self.mask_type,
            )
            self.feat_transformers.append(transformer)
            self.att_transformers.append(attention)

    def forward(self, x, prior=None):
        """Forward-pass of encoder.

        Args:
            x : input Tensor
            prior : mask for AttentiveTransformer

        Returns:
            sequence of outputs, regulariztion loss
        """
        x = self.initial_bn(x)

        bs = x.shape[0]  # batch size
        if prior is None:
            prior = torch.ones((bs, self.attention_dim)).to(x.device)

        M_loss = 0
        att = self.initial_splitter(x)[:, self.n_d :]
        steps_output = []
        for step in range(self.n_steps):
            M = self.att_transformers[step](prior, att)
            M_loss += torch.mean(torch.sum(torch.mul(M, torch.log(M + self.epsilon)), dim=1))
            # update prior
            prior = torch.mul(self.gamma - M, prior)
            # output
            M_feature_level = torch.matmul(M, self.group_attention_matrix.to(x.device))
            masked_x = torch.mul(M_feature_level, x)
            out = self.feat_transformers[step](masked_x)
            d = nn.ReLU()(out[:, : self.n_d])
            steps_output.append(d)
            # update attention
            att = out[:, self.n_d :]

        M_loss /= self.n_steps
        return steps_output, M_loss

    def forward_masks(self, x):
        """Magic forward-pass of encoder that returns masks.

        Args:
            x : input Tensor

        Returns:
            new and old masks.
        """
        x = self.initial_bn(x)
        bs = x.shape[0]  # batch size
        prior = torch.ones((bs, self.attention_dim)).to(x.device)
        M_explain = torch.zeros(x.shape).to(x.device)
        att = self.initial_splitter(x)[:, self.n_d :]
        masks = {}

        for step in range(self.n_steps):
            M = self.att_transformers[step](prior, att)
            M_feature_level = torch.matmul(M, self.group_attention_matrix.to(x.device))
            masks[step] = M_feature_level
            # update prior
            prior = torch.mul(self.gamma - M, prior)
            # output
            masked_x = torch.mul(M_feature_level, x)
            out = self.feat_transformers[step](masked_x)
            d = nn.ReLU()(out[:, : self.n_d])
            # explain
            step_importance = torch.sum(d, dim=1)
            M_explain += torch.mul(M_feature_level, step_importance.unsqueeze(dim=1))
            # update attention
            att = out[:, self.n_d :]

        return M_explain, masks


class FeatTransformer(torch.nn.Module):
    """Feature transformer from https://github.com/dreamquark-ai/tabnet.

    Args:
        input_dim : int
            Input size
        output_dim : int
            Output_size
        shared_layers : torch.nn.ModuleList
            The shared block that should be common to every step
        n_glu_independent : int
            Number of independent GLU layers
        virtual_batch_size : int
            Batch size for Ghost Batch Normalization within GLU block(s)
        momentum : float
            Float value between 0 and 1 which will be used for momentum in batch norm
    """

    def __init__(
        self,
        input_dim,
        output_dim,
        shared_layers,
        n_glu_independent,
        virtual_batch_size=128,
        momentum=0.02,
    ):
        super(FeatTransformer, self).__init__()
        params = {
            "n_glu": n_glu_independent,
            "virtual_batch_size": virtual_batch_size,
            "momentum": momentum,
        }

        if shared_layers is None:
            # no shared layers
            self.shared = torch.nn.Identity()
            is_first = True
        else:
            self.shared = GLU_Block(
                input_dim,
                output_dim,
                first=True,
                shared_layers=shared_layers,
                n_glu=len(shared_layers),
                virtual_batch_size=virtual_batch_size,
                momentum=momentum,
            )
            is_first = False

        if n_glu_independent == 0:
            # no independent layers
            self.specifics = torch.nn.Identity()
        else:
            spec_input_dim = input_dim if is_first else output_dim
            self.specifics = GLU_Block(spec_input_dim, output_dim, first=is_first, **params)

    def forward(self, x):
        """Forward-pass."""
        x = self.shared(x)
        x = self.specifics(x)
        return x


class GLU_Block(torch.nn.Module):
    """Independent GLU block, specific to each step.

    Code from https://github.com/dreamquark-ai/tabnet.

    Args:
        input_dim : int
            Input size
        output_dim : int
            Output_size
        shared_layers : torch.nn.ModuleList
            The shared block that should be common to every step
        n_glu : int
            Number of independent GLU layers
        virtual_batch_size : int
            Batch size for Ghost Batch Normalization within GLU block(s)
        momentum : float
            Float value between 0 and 1 which will be used for momentum in batch norm
        first : bool
            if the first layer of the block has no scale multiplication or not
    """

    def __init__(
        self,
        input_dim,
        output_dim,
        n_glu=2,
        first=False,
        shared_layers=None,
        virtual_batch_size=128,
        momentum=0.02,
    ):
        super(GLU_Block, self).__init__()
        self.first = first
        self.shared_layers = shared_layers
        self.n_glu = n_glu
        self.glu_layers = torch.nn.ModuleList()

        params = {"virtual_batch_size": virtual_batch_size, "momentum": momentum}

        fc = shared_layers[0] if shared_layers else None
        self.glu_layers.append(GLU_Layer(input_dim, output_dim, fc=fc, **params))
        for glu_id in range(1, self.n_glu):
            fc = shared_layers[glu_id] if shared_layers else None
            self.glu_layers.append(GLU_Layer(output_dim, output_dim, fc=fc, **params))

    def forward(self, x):
        """Forward-pass."""
        scale = torch.sqrt(torch.FloatTensor([0.5]).to(x.device))
        if self.first:  # the first layer of the block has no scale multiplication
            x = self.glu_layers[0](x)
            layers_left = range(1, self.n_glu)
        else:
            layers_left = range(self.n_glu)

        for glu_id in layers_left:
            x = torch.add(x, self.glu_layers[glu_id](x))
            x = x * scale
        return x


class GLU_Layer(torch.nn.Module):
    """GLU layer implementation.

    Args:
        input_dim : int
            Input size
        output_dim : int
            Output_size
        fc : torch.nn.Module
            Optional fully-connected layer
        virtual_batch_size : int
            Batch size for Ghost Batch Normalization within GLU block(s)
        momentum : float
            Float value between 0 and 1 which will be used for momentum in batch norm

    """

    def __init__(self, input_dim, output_dim, fc=None, virtual_batch_size=128, momentum=0.02):
        super(GLU_Layer, self).__init__()

        self.output_dim = output_dim
        if fc:
            self.fc = fc
        else:
            self.fc = nn.Linear(input_dim, 2 * output_dim, bias=False)
        _initialize_glu(self.fc, input_dim, 2 * output_dim)

        self.bn = GhostBatchNorm(2 * output_dim, virtual_batch_size=virtual_batch_size, momentum=momentum)

    def forward(self, x):
        """Forward-pass."""
        x = self.fc(x)
        x = self.bn(x)
        out = torch.mul(x[:, : self.output_dim], torch.sigmoid(x[:, self.output_dim :]))
        return out


class AttentiveTransformer(torch.nn.Module):
    """Attention transformer.

    Code from https://github.com/dreamquark-ai/tabnet.

    Args:
        input_dim : int
            Input size
        group_dim : int
            Number of groups for features
        virtual_batch_size : int
            Batch size for Ghost Batch Normalization
        momentum : float
            Float value between 0 and 1 which will be used for momentum in batch norm
        mask_type : str
            Either "sparsemax" or "entmax" : this is the masking function to use
    """

    def __init__(
        self,
        input_dim,
        group_dim,
        group_matrix,
        virtual_batch_size=128,
        momentum=0.02,
        mask_type="sparsemax",
    ):
        super(AttentiveTransformer, self).__init__()
        self.fc = nn.Linear(input_dim, group_dim, bias=False)
        _initialize_non_glu(self.fc, input_dim, group_dim)
        self.bn = GhostBatchNorm(group_dim, virtual_batch_size=virtual_batch_size, momentum=momentum)

        if mask_type == "sparsemax":
            # Sparsemax
            self.selector = Sparsemax()
        elif mask_type == "entmax":
            # Entmax
            self.selector = Entmax15()
        else:
            raise NotImplementedError("Please choose either sparsemax" + "or entmax as masktype")

    def forward(self, priors, processed_feat):
        """Forward-pass."""
        x = self.fc(processed_feat)
        x = self.bn(x)
        x = torch.mul(x, priors)
        x = self.selector(x)
        return x



"""utils for transformers."""

import numpy as np

from scipy.stats import mode


def get_mode(x):
    """Helper function to calculate mode."""
    return mode(x)[0][0]


class GroupByProcessor:
    """Helper class to calculate group_by features."""

    def __init__(self, keys):
        super().__init__()

        assert keys is not None

        self.index, self.keys_as_int = np.unique(keys, return_inverse=True)
        self.n_keys = max(self.keys_as_int) + 1
        self.set_indices()

    def set_indices(self):
        """Sets indices for keys."""
        self.indices = [[] for i in range(self.n_keys)]
        for i, k in enumerate(self.keys_as_int):
            self.indices[k].append(i)
        self.indices = [np.array(elt) for elt in self.indices]

    def apply(self, functions, vectors):
        """Applies functions to vectors."""
        assert functions is not None
        assert vectors is not None

        if isinstance(functions, list):
            return [[fun(vec[idx].tolist()) for fun, vec in zip(functions, vectors)] for idx in (self.indices)]
        else:
            return [functions(vectors[idx].tolist()) for idx in (self.indices)]


class GroupByFactory:
    """Factory to create group_by classes."""

    @staticmethod
    def get_GroupBy(kind):
        """Uses string identifiers to locate appropriate implementation.

        Args:
            kind: groupby feature transformation class

        Example:
            GroupByFactory.get_GroupBy('delta_mean')

        Returns:
            Object of GroupByBase implementing selected feature.

        Raises:
            ValueError: if identifier is not found.
        """
        assert kind is not None

        available_classes = [
            GroupByNumDeltaMean,
            GroupByNumDeltaMedian,
            GroupByNumMin,
            GroupByNumMax,
            GroupByNumStd,
            GroupByCatMode,
            GroupByCatIsMode,
        ]

        for class_name in available_classes:
            if kind == class_name.class_kind:
                return class_name(
                    class_name.class_kind,
                    class_name.class_fit_func,
                    class_name.class_transform_func,
                )

        raise ValueError(
            f"Unsupported kind: {kind}, available={[class_name.class_kind for class_name in available_classes]}"
        )
        return


class GroupByBase:
    """Base class for all group_by features.

    Note:
        Typically is created from GroupByFactory.

    Args:
            kind (string): Id of group_by feature.
            fit_func (function): function to calculate groups.
            transform_func (function): function to calculate statistics based on fitted groups.

    Example:
        GroupByBase(GroupByNumDeltaMean.class_kind, GroupByNumDeltaMean.class_fit_func, GroupByNumDeltaMean.class_transform_func)

    """

    def __init__(self, kind, fit_func, transform_func):
        super().__init__()

        self.kind = kind
        self.fit_func = fit_func
        self.transform_func = transform_func

        self._dict = None

    def get_dict(self):
        """Gets dict with features' statistics."""
        return self._dict

    def set_dict(self, dict):
        """Sets dict with features' statistics."""
        self._dict = dict

    def fit(self, data, group_by_processor, feature_column):
        """Calculate groups.

        Note:
            GroupByProcessor must be initialiaed before call to this function.

        Args:
            data (dataset): input data to extract ``feature_column``.
            group_by_processor (GroupByProcessor): processor, containing groups.
            feature_column (string): name of column to calculate statistics.

        Returns:
            self

        """
        assert data is not None
        assert group_by_processor is not None
        assert feature_column is not None

        assert self.fit_func is not None

        feature_values = data[feature_column].to_numpy()
        self._dict = dict(
            zip(
                group_by_processor.index,
                group_by_processor.apply(self.fit_func, feature_values),
            )
        )

        assert self._dict is not None

        return self

    def transform(self, data, value):
        """Calculate features statistics.

        Note:
            ``fit`` function must be called before ``transform``.

        Args:
            data (dataset): input data to extract ``value['group_column']`` and ``value['feature_column']``.
            value (dict): column names.

        Returns:
            transformed data

        """
        assert data is not None
        assert value is not None

        assert self.transform_func is not None

        group_values = data[value["group_column"]].to_numpy()
        feature_values = data[value["feature_column"]].to_numpy()
        result = self.transform_func(
            tuple(
                [
                    np.nan_to_num(np.array(np.vectorize(self._dict.get)(group_values), dtype=float)),
                    feature_values,
                ]
            )
        ).reshape(-1, 1)

        assert result is not None
        return result


class GroupByNumDeltaMean(GroupByBase):
    """Groupby delta mean class."""

    class_kind = "delta_mean"
    class_fit_func = np.nanmean

    @staticmethod
    def class_transform_func(values):
        """Get difference between feature values inside group and mean value."""
        return values[1] - values[0]


class GroupByNumDeltaMedian(GroupByBase):
    """Groupby delta median class."""

    class_kind = "delta_median"
    class_fit_func = np.nanmedian

    @staticmethod
    def class_transform_func(values):
        """Get difference between feature values inside group and median value."""
        return values[1] - values[0]


class GroupByNumMin(GroupByBase):
    """Groupby min class."""

    class_kind = "min"
    class_fit_func = np.nanmin

    @staticmethod
    def class_transform_func(values):
        """Get min value inside each group."""
        return values[0]


class GroupByNumMax(GroupByBase):
    """Groupby max class."""

    class_kind = "max"
    class_fit_func = np.nanmax

    @staticmethod
    def class_transform_func(values):
        """Get max value inside each group."""
        return values[0]


class GroupByNumStd(GroupByBase):
    """Groupby std class."""

    class_kind = "std"
    class_fit_func = np.nanstd

    @staticmethod
    def class_transform_func(values):
        """Get std value inside each group."""
        return values[0]


class GroupByCatMode(GroupByBase):
    """Groupby cat mode class."""

    class_kind = "mode"
    class_fit_func = get_mode

    @staticmethod
    def class_transform_func(values):
        """Get category mode inside each group."""
        return values[0]


class GroupByCatIsMode(GroupByBase):
    """Groupby is mode class."""

    class_kind = "is_mode"
    class_fit_func = get_mode

    @staticmethod
    def class_transform_func(values):
        """Check if category value is mode inside each group."""
        return values[0] == values[1]



"""Importance based selectors."""

from typing import Optional
from typing import TypeVar

from lightautoml.validation.base import TrainValidIterator

from ...dataset.base import LAMLDataset
from ...ml_algo.base import MLAlgo
from ..features.base import FeaturesPipeline
from .base import ImportanceEstimator
from .base import SelectionPipeline


ImportanceEstimatedAlgo = TypeVar("ImportanceEstimatedAlgo", bound=ImportanceEstimator)


class ModelBasedImportanceEstimator(ImportanceEstimator):
    """Base class for performing feature selection using model feature importances."""

    def fit(
        self,
        train_valid: Optional[TrainValidIterator] = None,
        ml_algo: Optional[ImportanceEstimatedAlgo] = None,
        preds: Optional[LAMLDataset] = None,
    ):
        """Find the importances of features.

        Args:
            train_valid: dataset iterator.
            ml_algo: ML algorithm used for importance estimation.
            preds: predicted target values.

        """
        assert (
            ml_algo is not None
        ), "ModelBasedImportanceEstimator: raw importances are None and no MLAlgo to calculate them."
        self.raw_importances = ml_algo.get_features_score()


class ImportanceCutoffSelector(SelectionPipeline):
    """Selector based on importance threshold.

    It is important that data which passed to ``.fit``
    should be ok to fit `ml_algo` or preprocessing pipeline should be defined.

    Args:
        feature_pipeline: Composition of feature transforms.
        ml_algo: Tuple (MlAlgo, ParamsTuner).
        imp_estimator: Feature importance estimator.
        fit_on_holdout: If use the holdout iterator.
        cutoff: Threshold to cut-off features.

    """

    def __init__(
        self,
        feature_pipeline: Optional[FeaturesPipeline],
        ml_algo: MLAlgo,
        imp_estimator: ImportanceEstimator,
        fit_on_holdout: bool = True,
        cutoff: float = 0.0,
    ):
        super().__init__(feature_pipeline, ml_algo, imp_estimator, fit_on_holdout)
        self.cutoff = cutoff

    def perform_selection(self, train_valid: Optional[TrainValidIterator] = None):
        """Select features based on cutoff value.

        Args:
            train_valid: Not used.

        """
        imp = self.imp_estimator.get_features_score()
        self.map_raw_feature_importances(imp)
        selected = self.mapped_importances.index.values[self.mapped_importances.values > self.cutoff]
        if len(selected) == 0:
            selected = self.mapped_importances.index.values[:1]
        self._selected_features = list(selected)



"""AutoML presets for image data."""

from lightautoml.utils.installation import __validate_extra_deps


__validate_extra_deps("cv", error=True)


import os

from typing import Optional
from typing import Sequence

import torch

from pandas import DataFrame

from ...ml_algo.boost_cb import BoostCB
from ...ml_algo.boost_lgbm import BoostLGBM
from ...ml_algo.linear_sklearn import LinearLBFGS
from ...ml_algo.tuning.optuna import OptunaTuner
from ...pipelines.features.base import FeaturesPipeline
from ...pipelines.features.image_pipeline import ImageAutoFeatures
from ...pipelines.features.image_pipeline import ImageSimpleFeatures
from ...pipelines.features.lgb_pipeline import LGBAdvancedPipeline
from ...pipelines.features.linear_pipeline import LinearFeatures
from ...pipelines.ml.nested_ml_pipe import NestedTabularMLPipeline
from ...pipelines.selection.base import SelectionPipeline
from ...reader.base import PandasToPandasReader
from ...reader.tabular_batch_generator import ReadableToDf
from ...tasks import Task
from ..blend import WeightedBlender
from .base import upd_params
from .tabular_presets import NumpyDataset
from .tabular_presets import TabularAutoML


_base_dir = os.path.dirname(__file__)
# set initial runtime rate guess for first level models
_time_scores = {"lgb": 1, "lgb_tuned": 3, "linear_l2": 0.7, "cb": 2, "cb_tuned": 6, "nn": 1, "rf": 5, "rf_tuned": 10}


# TODO: add text feature selection
class TabularCVAutoML(TabularAutoML):
    """Classic preset - work with tabular and image data.

    Supported data roles - numbers, dates, categories, images. Limitations - no memory management.
    GPU support in catboost/lightgbm (if installed for GPU).

    Commonly _params kwargs (ex. timing_params) set via config file (config_path argument).
    If you need to change just few params, it's possible to pass it as dict of dicts, like json.
    To get available params please look on default config template. Also you can find there param description.
    To generate config template call :func:`TabularCVAutoML.get_config('config_path.yml')`.

    Args:
        task: Task to solve.
        timeout: Timeout in seconds.
        memory_limit: Memory limit that are passed to each automl.
        cpu_limit: CPU limit that that are passed to each automl.
        gpu_ids: GPU IDs that are passed to each automl.
        timing_params: Timing param dict.
        config_path: Path to config file.
        general_params: General param dict
        reader_params: Reader param dict.
        read_csv_params: Params to pass :func:`pandas.read_csv`
            (case of train/predict from file).
        nested_cv_params: Param dict for nested cross-validation.
        tuning_params: Params of Optuna tuner.
        selection_params: Params of feature selection.
        lgb_params: Params of lightgbm model.
        cb_params: Params of catboost model.
        linear_l2_params: Params of linear model.
        gbm_pipeline_params: Params of feature generation
            for boosting models.
        linear_pipeline_params: Params of feature generation
            for linear models.
        cv_simple_features: Params of color histogram features.
        autocv_features: Params of image embeddings features.

    """

    _default_config_path = "image_config.yml"

    _time_scores = {"lgb": 1, "lgb_tuned": 3, "linear_l2": 0.7, "cb": 2, "cb_tuned": 6, "rf": 5, "rf_tuned": 10}

    def __init__(
        self,
        task: Task,
        timeout: int = 3600,
        memory_limit: int = 16,
        cpu_limit: int = 4,
        gpu_ids: Optional[str] = "all",
        timing_params: Optional[dict] = None,
        config_path: Optional[str] = None,
        general_params: Optional[dict] = None,
        reader_params: Optional[dict] = None,
        read_csv_params: Optional[dict] = None,
        nested_cv_params: Optional[dict] = None,
        tuning_params: Optional[dict] = None,
        selection_params: Optional[dict] = None,
        lgb_params: Optional[dict] = None,
        cb_params: Optional[dict] = None,
        rf_params: Optional[dict] = None,
        linear_l2_params: Optional[dict] = None,
        gbm_pipeline_params: Optional[dict] = None,
        linear_pipeline_params: Optional[dict] = None,
        cv_simple_features: Optional[dict] = None,
        autocv_features: Optional[dict] = None,
    ):
        super().__init__(
            task,
            timeout,
            memory_limit,
            cpu_limit,
            gpu_ids,
            timing_params,
            config_path,
        )

        # upd manual params
        for name, param in zip(
            [
                "general_params",
                "reader_params",
                "read_csv_params",
                "nested_cv_params",
                "tuning_params",
                "selection_params",
                "lgb_params",
                "cb_params",
                "rf_params",
                "linear_l2_params",
                "gbm_pipeline_params",
                "linear_pipeline_params",
                "cv_simple_features",
                "autocv_features",
            ],
            [
                general_params,
                reader_params,
                read_csv_params,
                nested_cv_params,
                tuning_params,
                selection_params,
                lgb_params,
                cb_params,
                rf_params,
                linear_l2_params,
                gbm_pipeline_params,
                linear_pipeline_params,
                cv_simple_features,
                autocv_features,
            ],
        ):
            if param is None:
                param = {}
            self.__dict__[name] = upd_params(self.__dict__[name], param)

    def infer_auto_params(self, train_data: DataFrame, multilevel_avail: bool = False):
        """Infer automatic parameters."""
        # infer gpu params
        gpu_cnt = torch.cuda.device_count()
        gpu_ids = self.gpu_ids
        if gpu_cnt > 0 and gpu_ids:
            if gpu_ids == "all":
                gpu_ids = ",".join(list(map(str, range(gpu_cnt))))

            self.autocv_features["device"] = gpu_ids.split(",")

            if self.general_params["use_algos"] == "auto":
                self.general_params["use_algos"] = [["linear_l2", "cb"]]

        else:
            self.autocv_features["device"] = "cpu"

            if self.general_params["use_algos"] == "auto":
                self.general_params["use_algos"] = [["linear_l2", "lgb"]]

        # check all n_jobs params
        cpu_cnt = min(os.cpu_count(), self.cpu_limit)
        torch.set_num_threads(cpu_cnt)

        if "n_jobs" in self.autocv_features:
            self.autocv_features["n_jobs"] = min(self.autocv_features["n_jobs"], cpu_cnt)
        else:
            self.autocv_features["n_jobs"] = cpu_cnt

        if "n_jobs" in self.cv_simple_features:
            self.cv_simple_features["n_jobs"] = min(self.cv_simple_features["n_jobs"], cpu_cnt)
        else:
            self.cv_simple_features["n_jobs"] = cpu_cnt

        # other params as tabular
        super().infer_auto_params(train_data, multilevel_avail)

    def get_cv_pipe(self, type: str = "simple") -> Optional[FeaturesPipeline]:
        """Get CV pipeline."""
        if type == "simple":
            return ImageSimpleFeatures(**self.cv_simple_features)
        elif type == "embed":
            return ImageAutoFeatures(**self.autocv_features)
        else:
            return None

    def get_linear(self, n_level: int = 1, pre_selector: Optional[SelectionPipeline] = None) -> NestedTabularMLPipeline:
        """Get linear pipeline."""
        # linear model with l2
        time_score = self.get_time_score(n_level, "linear_l2")
        linear_l2_timer = self.timer.get_task_timer("reg_l2", time_score)
        linear_l2_model = LinearLBFGS(timer=linear_l2_timer, **self.linear_l2_params)

        cv_l2_feats = self.get_cv_pipe(self.linear_pipeline_params["cv_features"])
        linear_l2_feats = LinearFeatures(output_categories=True, **self.linear_pipeline_params)
        if cv_l2_feats is not None:
            linear_l2_feats.append(cv_l2_feats)

        linear_l2_pipe = NestedTabularMLPipeline(
            [linear_l2_model],
            force_calc=True,
            pre_selection=pre_selector,
            features_pipeline=linear_l2_feats,
            **self.nested_cv_params
        )
        return linear_l2_pipe

    def get_gbms(
        self,
        keys: Sequence[str],
        n_level: int = 1,
        pre_selector: Optional[SelectionPipeline] = None,
    ):
        """Get gbm pipeline."""
        cv_gbm_feats = self.get_cv_pipe(self.gbm_pipeline_params["cv_features"])
        gbm_feats = LGBAdvancedPipeline(feats_imp=pre_selector, output_categories=False, **self.gbm_pipeline_params)
        if cv_gbm_feats is not None:
            gbm_feats.append(cv_gbm_feats)

        ml_algos = []
        force_calc = []
        for key, force in zip(keys, [True, False, False, False]):
            tuned = "_tuned" in key
            algo_key = key.split("_")[0]
            time_score = self.get_time_score(n_level, key)
            gbm_timer = self.timer.get_task_timer(algo_key, time_score)
            if algo_key == "lgb":
                gbm_model = BoostLGBM(timer=gbm_timer, **self.lgb_params)
            elif algo_key == "cb":
                gbm_model = BoostCB(timer=gbm_timer, **self.cb_params)
            else:
                raise ValueError("Wrong algo key")

            if tuned:
                gbm_tuner = OptunaTuner(
                    n_trials=self.tuning_params["max_tuning_iter"],
                    timeout=self.tuning_params["max_tuning_time"],
                    fit_on_holdout=self.tuning_params["fit_on_holdout"],
                )
                gbm_model = (gbm_model, gbm_tuner)
            ml_algos.append(gbm_model)
            force_calc.append(force)

        gbm_pipe = NestedTabularMLPipeline(
            ml_algos, force_calc, pre_selection=pre_selector, features_pipeline=gbm_feats, **self.nested_cv_params
        )

        return gbm_pipe

    def create_automl(self, **fit_args):
        """Create basic automl instance.

        See :meth:`lightautoml.automl.presets.base.AutoMLPreset.create_automl`.

        Args:
            **fit_args: params that are passed to ``.fit_predict`` method.

        """
        train_data = fit_args["train_data"]
        self.infer_auto_params(train_data)
        reader = PandasToPandasReader(task=self.task, **self.reader_params)

        pre_selector = self.get_selector()

        levels = []

        for n, names in enumerate(self.general_params["use_algos"]):
            lvl = []
            # regs
            if "linear_l2" in names:
                selector = None
                if "linear_l2" in self.selection_params["select_algos"] and (
                    self.general_params["skip_conn"] or n == 0
                ):
                    selector = pre_selector
                lvl.append(self.get_linear(n + 1, selector))

            gbm_models = [
                x for x in ["lgb", "lgb_tuned", "cb", "cb_tuned"] if x in names and x.split("_")[0] in self.task.losses
            ]

            if len(gbm_models) > 0:
                selector = None
                if "gbm" in self.selection_params["select_algos"] and (self.general_params["skip_conn"] or n == 0):
                    selector = pre_selector
                lvl.append(self.get_gbms(gbm_models, n + 1, selector))

            levels.append(lvl)

        # blend everything
        blender = WeightedBlender()

        # initialize
        self._initialize(reader, levels, skip_conn=self.general_params["skip_conn"], blender=blender, timer=self.timer)

    def predict(
        self,
        data: ReadableToDf,
        features_names: Optional[Sequence[str]] = None,
        batch_size: Optional[int] = None,
        n_jobs: Optional[int] = 1,
    ) -> NumpyDataset:
        """Get dataset with predictions.

        Almost same as :meth:`lightautoml.automl.base.AutoML.predict`
        on new dataset, with additional features.

        Additional features - working with different data formats.
        Supported now:

            - Path to `.csv`, `.parquet`, `.feather` files.
            - ``np.ndarray``, or dict of ``np.ndarray``. For example,
              ``{'data': X...}``. In this case roles are optional,
              but `train_features` and `valid_features` required.
            - ``pandas.DataFrame``.

        Parallel inference - you can pass ``n_jobs`` to speedup
        prediction (requires more RAM).
        Batch_inference - you can pass ``batch_size``
        to decrease RAM usage (may be longer).

        Args:
            data: Dataset to perform inference.
            features_names: Optional features names,
                if cannot be inferred from `train_data`.
            batch_size: Batch size or ``None``.
            n_jobs: Number of jobs.

        Returns:
            Dataset with predictions.

        """
        return super().predict(data, features_names, batch_size)



"""Base classes for MLPipeline."""

from typing import List
from typing import Optional
from typing import Sequence
from typing import Tuple
from typing import Union

from lightautoml.validation.base import TrainValidIterator

from ...dataset.base import LAMLDataset
from ...dataset.utils import concatenate
from ...ml_algo.base import MLAlgo
from ...ml_algo.tuning.base import DefaultTuner
from ...ml_algo.tuning.base import ParamsTuner
from ...ml_algo.utils import tune_and_fit_predict
from ..features.base import EmptyFeaturePipeline
from ..features.base import FeaturesPipeline
from ..selection.base import EmptySelector
from ..selection.base import SelectionPipeline


class MLPipeline:
    """Single ML pipeline.

    Merge together stage of building ML model
    (every step, excluding model training, is optional):

        - Pre selection: select features from input data.
          Performed by
          :class:`~lightautoml.pipelines.features.SelectionPipeline`.
        - Features generation: build new features from selected.
          Performed by
          :class:`~lightautoml.pipelines.features.FeaturesPipeline`.
        - Post selection: One more selection step - from created features.
          Performed by
          :class:`~lightautoml.pipelines.features.SelectionPipeline`.
        - Hyperparams optimization for one or multiple ML models.
          Performed by
          :class:`~lightautoml.ml_algo.tuning.base.ParamsTuner`.
        - Train one or multiple ML models:
          Performed by :class:`~lightautoml.ml_algo.base.MLAlgo`.
          This step is the only required for at least 1 model.

    Args:
        ml_algos: Sequence of MLAlgo's or Pair - (MlAlgo, ParamsTuner).
        force_calc: Flag if single fold of ml_algo
            should be calculated anyway.
        pre_selection: Initial feature selection.
            If ``None`` there is no initial selection.
        features_pipeline: Composition of feature transforms.
        post_selection: Post feature selection.
            If ``None`` there is no post selection.

    """

    @property
    def used_features(self) -> List[str]:  # noqa D102
        return self.pre_selection.selected_features

    def __init__(
        self,
        ml_algos: Sequence[Union[MLAlgo, Tuple[MLAlgo, ParamsTuner]]],
        force_calc: Union[bool, Sequence[bool]] = True,
        pre_selection: Optional[SelectionPipeline] = None,
        features_pipeline: Optional[FeaturesPipeline] = None,
        post_selection: Optional[SelectionPipeline] = None,
    ):
        if pre_selection is None:
            pre_selection = EmptySelector()

        self.pre_selection = pre_selection

        if features_pipeline is None:
            features_pipeline = EmptyFeaturePipeline()

        self.features_pipeline = features_pipeline

        if post_selection is None:
            post_selection = EmptySelector()

        self.post_selection = post_selection

        self._ml_algos = []
        self.params_tuners = []

        for n, mt_pair in enumerate(ml_algos):

            try:
                # case when model and tuner are defined
                mod, tuner = mt_pair
            except (TypeError, ValueError):
                # case when only model is defined
                mod, tuner = mt_pair, DefaultTuner()

            mod.set_prefix("Mod_{0}".format(n))

            self._ml_algos.append(mod)
            self.params_tuners.append(tuner)

        self.force_calc = [force_calc] * len(self._ml_algos) if type(force_calc) is bool else force_calc
        # TODO: Do we need this assert?
        assert any(self.force_calc), "At least single algo in pipe should be forced to calc"

    def fit_predict(self, train_valid: TrainValidIterator) -> LAMLDataset:
        """Fit on train/valid iterator and transform on validation part.

        Args:
            train_valid: Dataset iterator.

        Returns:
            Dataset with predictions of all models.

        """
        self.ml_algos = []
        # train and apply pre selection
        train_valid = train_valid.apply_selector(self.pre_selection)

        # apply features pipeline
        train_valid = train_valid.apply_feature_pipeline(self.features_pipeline)

        # train and apply post selection
        train_valid = train_valid.apply_selector(self.post_selection)

        predictions = []

        for ml_algo, param_tuner, force_calc in zip(self._ml_algos, self.params_tuners, self.force_calc):
            ml_algo.debug = getattr(self, "debug", False)
            ml_algo, preds = tune_and_fit_predict(ml_algo, param_tuner, train_valid, force_calc)
            if ml_algo is not None:
                self.ml_algos.append(ml_algo)

                predictions.append(preds)

        assert (
            len(predictions) > 0
        ), "Pipeline finished with 0 models for some reason.\nProbably one or more models failed"

        predictions = concatenate(predictions)

        del self._ml_algos
        return predictions

    def predict(self, dataset: LAMLDataset) -> LAMLDataset:
        """Predict on new dataset.

        Args:
            dataset: Dataset used for prediction.

        Returns:
            Dataset with predictions of all trained models.

        """
        dataset = self.pre_selection.select(dataset)
        dataset = self.features_pipeline.transform(dataset)
        dataset = self.post_selection.select(dataset)

        predictions = []

        for model in self.ml_algos:
            pred = model.predict(dataset)
            predictions.append(pred)

        predictions = concatenate(predictions)

        return predictions

    def upd_model_names(self, prefix: str):
        """Update prefix pipeline models names.

        Used to fit inside AutoML where multiple models
        with same names may be trained.

        Args:
            prefix: New prefix name.

        """
        for n, mod in enumerate(self._ml_algos):
            mod.set_prefix(prefix)

    def prune_algos(self, idx: Sequence[int]):
        """Prune model from pipeline.

        Used to fit blender - some models may be excluded from final ensemble.

        Args:
            idx: Selected algos.

        """
        self.ml_algos = [x for (n, x) in enumerate(self.ml_algos) if n not in idx]



"""Tools for model training."""

import logging
import traceback

from typing import Optional
from typing import Tuple

import torch.nn as nn

from torch import Tensor
from torch import finfo

from ..dataset.base import LAMLDataset
from ..validation.base import TrainValidIterator
from .base import MLAlgo
from .tuning.base import ParamsTuner


logger = logging.getLogger(__name__)


def tune_and_fit_predict(
    ml_algo: MLAlgo,
    params_tuner: ParamsTuner,
    train_valid: TrainValidIterator,
    force_calc: bool = True,
) -> Tuple[Optional[MLAlgo], Optional[LAMLDataset]]:
    """Tune new algorithm, fit on data and return algo and predictions.

    Args:
        ml_algo: ML algorithm that will be tuned.
        params_tuner: Tuner object.
        train_valid: Classic cv-iterator.
        force_calc: Flag if single fold of ml_algo should be calculated anyway.

    Returns:
        Tuple (BestMlAlgo, predictions).

    """
    timer = ml_algo.timer
    timer.start()
    single_fold_time = timer.estimate_folds_time(1)

    # if force_calc is False we check if it make sense to continue
    if not force_calc and (
        (single_fold_time is not None and single_fold_time > timer.time_left) or timer.time_limit_exceeded()
    ):
        return None, None

    if params_tuner.best_params is None:
        # this try/except clause was added because catboost died for some unexpected reason
        if getattr(ml_algo, "debug", False):
            new_algo, preds = params_tuner.fit(ml_algo, train_valid)
        else:
            try:
                # TODO: Set some conditions to the tuner
                new_algo, preds = params_tuner.fit(ml_algo, train_valid)
            except Exception as e:
                logger.info2("Model {0} failed during params_tuner.fit call.\n\n{1}".format(ml_algo.name, e))
                return None, None

        if preds is not None:
            return new_algo, preds

    if not force_calc and (
        (single_fold_time is not None and single_fold_time > timer.time_left) or timer.time_limit_exceeded()
    ):
        return None, None

    ml_algo.params = params_tuner.best_params
    # this try/except clause was added because catboost died for some unexpected reason
    if getattr(ml_algo, "debug", False):
        preds = ml_algo.fit_predict(train_valid)
    else:
        try:
            preds = ml_algo.fit_predict(train_valid)
        except Exception as e:
            logger.info2("Model {0} failed during ml_algo.fit_predict call.\n\n{1}".format(ml_algo.name, e))
            logger.info3(traceback.format_exc())
            return None, None

    return ml_algo, preds


class SoftmaxClip(nn.Module):
    """Softmax with clip-norm.

    Args:
        dim : A dimension along which Softmax will be computed (so every slice
            along dim will sum to 1).
    """

    def __init__(self, dim: Optional[int] = None) -> None:
        super(SoftmaxClip, self).__init__()
        self.dim = dim
        self.smax = nn.Softmax(dim=dim)

    def forward(self, inputs: Tensor) -> Tensor:
        """Inference phase.

        Args:
            inputs: data to softmax and clip.

        Returns:
            transformed values.

        """
        inputs = self.smax(inputs)
        eps = 2 * finfo(inputs.dtype).eps
        inputs = inputs.clip(eps, 1 - eps)
        inputs /= inputs.sum(dim=self.dim)[:, None]
        return inputs



"""Whitebox AutoML presets."""

import logging
import os

from copy import deepcopy
from typing import Any
from typing import Iterable
from typing import Optional
from typing import Sequence
from typing import cast

from ...dataset.np_pd_dataset import NumpyDataset
from ...ml_algo.whitebox import WbMLAlgo
from ...pipelines.ml.whitebox_ml_pipe import WBPipeline
from ...reader.base import PandasToPandasReader
from ...tasks import Task
from .base import AutoMLPreset
from .base import upd_params


logger = logging.getLogger(__name__)
_base_dir = os.path.dirname(__file__)


class WhiteBoxPreset(AutoMLPreset):
    """Preset for AutoWoE - logistic regression over binned features (scorecard).

    Supported data roles - numbers, dates, categories.

    Limitations:

        - Simple time management.
        - No memory management.
        - Working only with :class:`pandas.DataFrame`.
        - No batch inference.
        - No text support.
        - No parallel execution.
        - No batch inference.
        - No GPU usage.
        - No cross-validation scheme. Supports only holdout validation
          (cv is created inside AutoWoE, but no oof pred returned).

    Common usecase - fit lightweight interpretable model
    for binary classification task.

    Commonly _params kwargs (ex. timing_params) set via
    config file (config_path argument).
    If you need to change just few params, it's possible to pass it
    as dict of dicts, like json.
    To get available params please look on default config template.
    Also you can find there param description
    To generate config template call
    :func:`WhiteBoxPreset.get_config('config_path.yml')`.

    Args:
        task: Task to solve.
        timeout: Timeout in seconds.
        memory_limit: Memory limit that are passed to each automl.
        cpu_limit: CPU limit that that are passed to each automl.
        gpu_ids: GPU IDs that are passed to each automl.
        timing_params: Timing param dict.
        config_path: Path to config file.
        general_params: General param dict.
        reader_params: Reader param dict.
        read_csv_params: Params to pass :func:`pandas.read_csv`
            (case of train/predict from file).
        whitebox_params: Params of WhiteBox algo (look at config file).

    """

    _default_config_path = "whitebox_config.yml"

    @property
    def whitebox(self):
        """Get wrapped AutoWoE object.

        Returns:
            Model.

        """
        return self.levels[0][0].whitebox

    def __init__(
        self,
        task: Task,
        timeout: int = 3600,
        memory_limit: int = 16,
        cpu_limit: int = 4,
        gpu_ids: Optional[str] = None,
        timing_params: Optional[dict] = None,
        config_path: Optional[str] = None,
        general_params: Optional[dict] = None,
        reader_params: Optional[dict] = None,
        read_csv_params: Optional[dict] = None,
        whitebox_params: Optional[dict] = None,
    ):
        super().__init__(
            task,
            timeout,
            memory_limit,
            cpu_limit,
            gpu_ids,
            timing_params,
            config_path,
        )

        # upd manual params
        for name, param in zip(
            [
                "general_params",
                "reader_params",
                "read_csv_params",
                "whitebox_params",
            ],
            [
                general_params,
                reader_params,
                read_csv_params,
                whitebox_params,
            ],
        ):
            if param is None:
                param = {}
            self.__dict__[name] = upd_params(self.__dict__[name], param)

    def infer_auto_params(self, **kwargs):
        """Infer automatic parameters."""
        # check all n_jobs params
        cpu_cnt = min(os.cpu_count(), self.cpu_limit)
        self.whitebox_params["default_params"]["n_jobs"] = min(
            self.whitebox_params["default_params"]["n_jobs"], cpu_cnt
        )
        self.reader_params["n_jobs"] = min(self.reader_params["n_jobs"], cpu_cnt)

    def create_automl(self, *args, **kwargs):
        """Create basic :class:`~lightautoml.automl.presets.whitebox_presets.WhiteBoxPreset` instance from data.

        Args:
            *args: Not used.
            **kwargs: everything passed to ``.fit_predict``.

        """
        self.infer_auto_params()
        reader = PandasToPandasReader(task=self.task, **self.reader_params)
        wb_timer = self.timer.get_task_timer("wb", 1)

        whitebox_params = deepcopy(self.whitebox_params)
        whitebox_params["fit_params"] = self.fit_params
        whitebox_params["report"] = self.general_params["report"]
        whitebox_params["verbose"] = self.verbose

        _whitebox = WbMLAlgo(timer=wb_timer, default_params=whitebox_params)
        whitebox = WBPipeline(_whitebox)
        levels = [[whitebox]]

        # initialize
        self._initialize(reader, levels, skip_conn=False, timer=self.timer)

    def fit_predict(
        self,
        train_data: Any,
        roles: dict,
        train_features: Optional[Sequence[str]] = None,
        cv_iter: Optional[Iterable] = None,
        valid_data: Optional[Any] = None,
        valid_features: Optional[Sequence[str]] = None,
        verbose: int = 0,
        **fit_params
    ) -> NumpyDataset:
        """Fit and get prediction on validation dataset.

        Almost same as :meth:`lightautoml.automl.base.AutoML.fit_predict`.

        Additional features - working with different data formats.
        Supported now:

            - Path to ``.csv``, ``.parquet``, ``.feather`` files.
            - :class:`~numpy.ndarray`, or dict of :class:`~numpy.ndarray`.
              For example, ``{'data': X...}``. In this case,
              roles are optional, but `train_features`
              and `valid_features` required.
            - :class:`pandas.DataFrame`.

        Args:
            train_data: Dataset to train.
            roles: Roles dict.
            train_features: Optional features names, if can't be inferred from `train_data`.
            cv_iter: Custom cv-iterator. For example, :class:`~lightautoml.validation.np_iterators.TimeSeriesIterator`.
            valid_data: Optional validation dataset.
            valid_features: Optional validation dataset features if cannot be inferred from `valid_data`.
            verbose: Controls the verbosity: the higher, the more messages.
                <1  : messages are not displayed;
                >=1 : the computation process for layers is displayed;
                >=2 : the information about folds processing is also displayed;
                >=3 : the hyperparameters optimization process is also displayed;
                >=4 : the training process for every algorithm is displayed;
            fit_params: Others.

        Returns:
            Dataset with predictions. Call ``.data`` to get predictions array.

        """
        assert cv_iter is None or len(cv_iter) == 2, "Expect custom iterator with len 2"
        if valid_data is None and cv_iter is None:
            logger.info2("Validation data is not set. Train will be used as valid in report and valid prediction")
            valid_data = train_data
            valid_features = train_features

        self.fit_params = fit_params
        self.verbose = verbose

        pred = super().fit_predict(train_data, roles, train_features, cv_iter, valid_data, valid_features, verbose)

        return cast(NumpyDataset, pred)

    def predict(
        self,
        data: Any,
        features_names: Optional[Sequence[str]] = None,
        report: bool = False,
    ) -> NumpyDataset:
        """Almost same as AutoML ``.predict`` with additional features.

        Additional features - generate extended WhiteBox
        ``report=True`` passed to args.

        Args:
            data: Dataset to perform inference.
            features_names: Optional features names, if can't be inferred from `train_data`.
            report: Flag if we need inner WhiteBox report update (``True`` is slow). Only if ``general_params['report'] = True``.

        Returns:
            Dataset with predictions.

        """
        dataset = self.reader.read(data, features_names, add_array_attrs=report)
        pred = self.levels[0][0].predict(dataset, report=report)

        return cast(NumpyDataset, pred)

    def create_model_str_desc(self) -> str:
        """String description of model."""
        return ""



"""Metrics and loss functions for Catboost."""

import logging

from typing import Callable
from typing import Dict
from typing import Optional
from typing import Union

import numpy as np

from .base import Loss
from .base import fw_rmsle


logger = logging.getLogger(__name__)


def bw_clipping(x):
    """Clip to [0, 1]."""
    return np.clip(x, 0, 1)


def cb_str_loss_wrapper(name: str, **params: Optional[Dict]) -> str:
    """CatBoost loss name wrapper, if it has keyword args.  # noqa D403

    Args:
        name: One of CatBoost loss names.
        **params: Additional parameters.

    Returns:
        Wrapped CatBoost loss name.

    """
    return name + ":" + ";".join([k + "=" + str(v) for (k, v) in params.items()])


_cb_loss_mapping = {
    "mse": ("RMSE", None, None),
    "mae": ("MAE", None, None),
    "logloss": ("Logloss", None, None),
    "rmsle": ("RMSE", fw_rmsle, np.expm1),
    "mape": ("MAPE", None, None),
    "quantile": ("Quantile", None, None),
    "fair": ("FairLoss", None, None),
    "huber": ("Huber", None, None),
    "crossentropy": ("MultiClass", None, None),
}

_cb_loss_params_mapping = {
    "quantile": {"q": "alpha"},
    "huber": {"a": "delta"},
    "fair": {"c": "smoothness"},
}

_cb_binary_metrics_dict = {
    "auc": "AUC",
    "logloss": "Logloss",
    "accuracy": "Accuracy",
}

_cb_reg_metrics_dict = {
    "mse": "RMSE",
    "mae": "MAE",
    "r2": "R2",
    "rmsle": "MSLE",
    "mape": "MAPE",
    "quantile": "Quantile",
    "fair": "FairLoss",
    "huber": "Huber",
}

_cb_multiclass_metrics_dict = {
    "auc": "AUC:type=Mu",  # for overfitting detector
    "auc_mu": "AUC:type=Mu",
    "accuracy": "Accuracy",
    "crossentropy": "MultiClass",
    "f1_macro": "TotalF1:average=Macro",
    "f1_micro": "TotalF1:average=Micro",
    "f1_weighted": "TotalF1:average=Weighted",
}
_cb_multireg_metric_dict = {
    "rmse": "MultiRMSE",
    "mse": "MultiRMSE",
    "mae": "MultiRMSE",
}

_cb_multilabel_metric_dict = {"logloss": "MultiCrossEntropy"}

_cb_metrics_dict = {
    "binary": _cb_binary_metrics_dict,
    "reg": _cb_reg_metrics_dict,
    "multiclass": _cb_multiclass_metrics_dict,
    "multi:reg": _cb_multireg_metric_dict,
    "multilabel": _cb_multilabel_metric_dict,
}


_cb_metric_params_mapping = {
    "quantile": {"q": "alpha"},
    "huber": {"a": "delta"},
    "fair": {"c": "smoothness"},
}


class CBLoss(Loss):
    """Loss used for CatBoost.

    Args:
        loss: String with one of default losses.
        loss_params: additional loss parameters.
            Format like in :mod:`lightautoml.tasks.custom_metrics`.
        fw_func: Forward transformation.
            Used for transformation of target and item weights.
        bw_func: Backward transformation.
            Used for predict values transformation.

    """

    def __init__(
        self,
        loss: Union[str, Callable],
        loss_params: Optional[Dict] = None,
        fw_func: Optional[Callable] = None,
        bw_func: Optional[Callable] = None,
    ):
        self.loss_params = {}
        if loss_params is not None:
            self.loss_params = loss_params

        if type(loss) is str:
            if loss in _cb_loss_mapping:
                loss_name, fw_func, bw_func = _cb_loss_mapping[loss]
                if loss in _cb_loss_params_mapping:
                    mapped_params = {_cb_loss_params_mapping[loss][k]: v for (k, v) in self.loss_params.items()}
                    self.fobj = None
                    self.fobj_name = cb_str_loss_wrapper(loss_name, **mapped_params)

                else:
                    self.fobj = None
                    self.fobj_name = loss_name
            else:
                raise ValueError("Unexpected loss for catboost")
                # special loss for catboost, that is not defined in _cb_loss_mapping
                # self.fobj = None
                # self.fobj_name = loss
        else:
            # custom catboost objective
            self.fobj = loss
            self.fobj_name = None

        if fw_func is not None:
            self._fw_func = fw_func

        if bw_func is not None:
            self._bw_func = bw_func

        self.fobj_params = {}
        if loss_params is not None:
            self.fobj_params = loss_params

        self.metric = None
        self.metric_name = None

    def set_callback_metric(
        self,
        metric: Union[str, Callable],
        greater_is_better: Optional[bool] = None,
        metric_params: Optional[Dict] = None,
        task_name: str = None,
    ):
        """Callback metric setter.

        Args:
            metric: Callback metric.
            greater_is_better: Whether or not higher value is better.
            metric_params: Additional metric parameters.
            task_name: Name of task. For now it omitted.

        """
        # TODO: for what cb_utils
        # How to say that this metric is special class if there any task type?

        assert task_name in ["binary", "reg", "multiclass", "multi:reg", "multilabel"], "Unknown task name: {}".format(
            task_name
        )

        self.metric_params = {}
        if metric_params is not None:
            self.metric_params = metric_params

        if type(metric) is str:
            self.metric = None
            _metric_dict = _cb_metrics_dict[task_name]
            if task_name == "multi:reg":
                logger.info2("CatBoost supports only MultiRMSE metric and loss for multi:reg task.")
                self.fobj = None
                self.fobj_name = "MultiRMSE"
            if task_name == "multilabel":
                logger.info2("CatBoost uses as obj. MultiCrossEntropy.")
                self.fobj = None
                self.fobj_name = "MultiCrossEntropy"

            if metric in _cb_metric_params_mapping:
                metric_params = {_cb_metric_params_mapping[metric][k]: v for (k, v) in self.metric_params.items()}
                self.metric_name = cb_str_loss_wrapper(_metric_dict[metric], **metric_params)
            else:
                self.metric_name = _metric_dict[metric]

        else:
            # TODO: Check it later
            self.metric_name = self.fobj_name
            self.metric_params = self.fobj_params
            self.metric = None

            if task_name == "multi:reg":
                logger.info2("CatBoost supports only MultiRMSE metric and loss for multi:reg task.")
                self.fobj = None
                self.fobj_name = "MultiRMSE"
                self.metric_name = "MultiRMSE"
            if task_name == "multilabel":
                logger.info2("CatBoost uses as obj. MultiCrossEntropy.")
                self.fobj = None
                self.fobj_name = "MultiCrossEntropy"
                self.metric_name = "MultiCrossEntropy"
